(************************************NOTICE***********************************
 * This source code and its compiled binary code are Copyright  (C) 1994  by *
 * The Eclectic Company, 1911 Packard, Ann Arbor, MI 48104-4733.  License is *
 * hereby granted to copy,  distribute,  and use them freely for non-commer- *
 * cial research purposes only,  provided:  (1) that this notice is included *
 * in the source code, and (2) that the "trademark" procedure is included in *
 * the compiled binary code.  Modifications and  recompilations of this code *
 * may also be performed freely for the purposes of non-commercial research, *
 * under the same conditions;  the author would  appreciate receiving a copy *
 * of any such modifications.                                                *
 *****************************************************************************)

{
 Automatic program to update data files and indices
 after changes have been made to the MONOSYL.DAT file.

 Written by: John M. Lawler
             Program in Linguistics
             University of Michigan
             Ann Arbor, MI 48109

 Ported to PC-DOS 9/28/87
 Added rime indexing by dynamic lists 10/29/87
 Automated 8/24/88
 DOS directory search and sigfile added 8/29/88
 Last compiled 8/10/95
}

{$R-}    {Range checking off}
{$B+}    {Boolean complete evaluation on}
(* {$S+}    {Stack checking on} *)
{$I+}    {I/O checking on}
(* {$N-}    {No numeric coprocessor}  *)
(* {$M 65500,16384,655360} {Turbo 3 default stack and heap} *)

(* Uses Crt; *)

const
     spellength = 12;
     asslength  = 4;
     rimlength  = 6;
     suflength  = 3;
     wordlength = 25;
     size       = 5500;
     blank      = ' ';
     prompt_y   = 24;
     status_y   = 25;
     dataname   = 'MONOSYL.DAT';
     signame    = 'MONOSYL.INI';

type
     wnumber   = 0..size;
     whatsit   = (ass, rim, suf);
     wordo     = string[wordlength];
     spelling  = string [spellength];
     assonance = string [asslength];
     rime      = string[rimlength];
     suffix    = string[suflength];
     screenline = string[82];
     entry_ptr = ^entry;
     entry     = record
                    spelling : spelling;
                    assonance: assonance;
                    rime     : rime;
                    suffix   : suffix;
                 end; {record entry definition}
     list_ptr   = ^list;
     list       = record
            no     : wnumber;
            next   : list_ptr;
                 end;
     index      = record
            num    : integer;
            name   : rime;
            list   : list_ptr;
                 end;

const
     status_line : screenline =
     ' Total                                                            ';
var
     i,
     j,
     k,
     thisnum,
     screenlin,
     total,
     count,
     which,
     count_x,
     status_x,
     number    : integer;
     c         : char;
     newheap,
     heaptop   :^integer;
     first,
     trademarked,
     already,
     changes,
     ok        : boolean;
     thiscard  : entry;
     location,
     cc        : screenline;
     thisentry : array [1..size] of entry_ptr;
     list_array: array [1..500] of index;
     kindof,
     filename,
     thisline  : wordo;
     prompt_string : screenline;
     itsa      : whatsit;
     sigfile,
     textfile  : text;
     fil       : file of entry;
     ndex      : file of integer;

     intcount  : array [1..100] of integer;
     intarray  : array [1..size] of integer;
     asslst    : array [1..100] of assonance;
     suflst    : array [1..25] of suffix;

{BEGIN screen routines}

procedure reversi (counter : integer);
begin
     textbackground (white);
     textcolor (black);
     write (counter);
     textbackground (black);
     textcolor (white);
end; {procedure reversv}

procedure reversv (text : screenline);
begin
     textbackground (white);
     textcolor (black);
     write (text);
     textbackground (black);
     textcolor (white);
end; {procedure reversv}

procedure trademark;
begin
     trademarked := true;
     clrscr;
     write ('     ');
     reversv ('               Monosyllable Indexer  *  Version 2.6               ');
     writeln;
     write ('     ');
     reversv ('             Copyright (C) 1995  The Eclectic Company             ');
     writeln;
     first := true;
end; {procedure trademark}

procedure wipescreen;
var i : integer;
begin
     clreol;
     for i := wherey+1 to 23 do
       begin
          gotoxy (1,i);
          clreol;
       end;
end;

procedure status (this : whatsit; no : integer);
var x : screenline;
    ix,
    l : integer;
begin
     case this of
          ass : if first = true then
                  begin
                    status_x := 9;
                    str (no, x);
                    x := 'Words: ' + x;
                    first := false;
                  end
                else
                  begin
                    status_x := 23;
                    str (no, x);
                    x := 'Assonances: ' + x;
                  end;
          rim : begin
                    status_x := 41;
                    str (no, x);
                    x := 'Rimes: ' + x;
                  end;
          suf : begin
                    status_x := 54;
                    str (no, x);
                    x := 'Suffixes: ' + x;
                  end;
           end; {case}
     l := length (x);
     for ix := status_x to status_x+l-1 do
        status_line[ix] := x[ix-status_x+1];
     gotoxy (6,status_y);
     reversv (status_line);
end;

procedure prompt (x : screenline);
const blank_line =
'                                                                  ';
begin
     gotoxy (6,prompt_y);
     textbackground (white);
     textcolor (black);
     write (blank_line);
     gotoxy (7,prompt_y);
     write (x);
     count_x := wherex;
     textbackground (black);
     textcolor (white);
     prompt_string := x;
end; {procedure prompt}

{END screen routines}

procedure load_data;

{internal} procedure parse (x : wordo; var y :entry);
     {converts line entry into record}
begin
     i := 1;
     while length (x) < wordlength do
       x := x + blank;  {pad entry out to appropriate length}
     with y do
       begin   {note that record entries are NOT shaved}
         spelling := copy (x, i, spellength);
         i := i + spellength;
         assonance := copy (x, i, asslength);
         i := i + asslength;
         rime := copy (x, i, rimlength);
         i := i + rimlength;
         suffix := copy (x, i, suflength);
       end; {with statement}
end; {internal procedure parse}

function get_init : screenline;
var temp_string : screenline;
    leng        : integer;
begin
     getdir (0, temp_string);      {get the default directory}
     assign (sigfile, signame);    {look for initialization file}
{$I-}
     reset (sigfile);
{$I+}
     if ioresult = 0 then          {if found...}
        begin
           readln (sigfile, temp_string);    {get the directory}
           close (sigfile);
        end
     else
        begin
          assign (sigfile, '\'+signame);     {if not, look in root directory}
{$I-}
          reset (sigfile);
{$I+}
          if ioresult = 0 then               {if it's there...}
             begin
                readln (sigfile, temp_string);  {get the directory}
                close (sigfile);
             end; {if}
        end; {else}                      
     leng := length (temp_string);
     if not (temp_string[leng] in ['\', ':']) then
        temp_string := temp_string + '\';
     get_init := temp_string;           {return directory}
end; {screenline function get_init}

var leng : integer;
    query_string : screenline;

begin {procedure load_data proper}
     location := get_init;
     ok := false;
     while not ok do
       begin
         assign (textfile, location+dataname);
{$I-}
         reset (textfile);
{$I+}
         if ioresult <> 0 then
          begin
             prompt_string :=
     location + dataname + ' not found. ';
             prompt (prompt_string);
             query_string :=
     ' What drive is '+dataname+ ' on? [default = this one]   '^H^H^H;
             gotoxy (6,5);
             reversv (query_string);
             read (kbd, c);
             if c = ^M then
                begin
                   getdir (0, cc);
                   c := cc[1];
                end;
             c := upcase (c);
             if c = 'Q' then halt;
             query_string :=
     ' What directory is '+dataname+' in?   ("Q" to quit) ';
             gotoxy (6,7);
             reversv (query_string);
             query_string :=  ' ' + c +
     ':                                                ';
             gotoxy (6,8);
             reversv (query_string);
             gotoxy (9,8);
             readln (cc);
             leng := length (cc);
             for i := 1 to leng do
                cc[i] := upcase (cc[i]);
             if cc = 'Q' then halt;
             if cc[leng] <> '\' then
                cc := cc + '\';
             location := c + ':' + cc;
             gotoxy (6,3);
             wipescreen;
             assign (sigfile, signame);    {write a new init file here}
             rewrite (sigfile);
             writeln (sigfile, location);
             close (sigfile);
          end
        else
          ok := true;
       end; {while not ok}
     mark (heaptop);
     prompt ('Loading Word #');
     number := 0;
     while not eof(textfile) do
       begin
          number := number + 1;
          readln (textfile,thisline);
          new (thisentry[number]);
          parse (thisline, thisentry[number]^);
          if (number mod 32 = 0) then
            begin
               gotoxy (count_x, prompt_y);
               reversi (number);
            end; {if}
       end; {while}
     status (ass, number);
     close (textfile);
end; {procedure load_data}

{BEGIN indexing routines}

procedure update;  {called by all indexing routines for screen display}
var
     i1,
     i2 : integer;
     s1,
     s2,
     s3 : wordo;
begin
     with thisentry[i+1]^ do
       begin
          s3 := spelling;
          count := count + 1;
          i1 := count;
          i2 := 12;
          case itsa of
          suf  : begin
               intcount[count] := 1;
               suflst[count] := suffix;
               s1 := 'suffix';
               s2 := suffix;
                 end; {case of suf}
          rim  : begin
               s1 := 'rime';
               s2 := rime;
               with list_array[i1] do
                  begin
                     num  := 1;
                     name := s2;
                     new (list);
                     list^.no := i;
                  end; {if and with}
                 end; {case of rim}
          ass  : begin
               asslst[count] := assonance;
               intcount[count] := 1;
               s1 := 'assonance';
               s2 := assonance;
               i2 := 8;
                 end; {case of ass}
          end; {case}
       end; {with}
     if screenlin = 0 then
        begin
           screenlin := screenlin + 1;
           gotoxy (1, screenlin+4);
           wipescreen;
        end;
     gotoxy (i2, screenlin+4);
     clreol;
     write ('Found '+s1+' no. ',i1:3,': '+s2+' in '+s3+' [Word #',i+1,']');
     screenlin := (screenlin + 1) mod 18;
end; {procedure update}

procedure make_assonance_list;  {for assonances only - no indexing done
                                 since data is already sorted for assonances}
begin
     itsa := ass;
     screenlin := 0;
     prompt ('Making list of assonances');
     count := 0;
     for i:= 0 to number-1 do
       begin
          with thisentry[i+1]^ do
               if count = 0 then
                    update
               else
                 begin
                    j := count;
                    already := false;
                    while (j >= 1) and (not already) do
                       if assonance = asslst[j] then
                         begin
                            already := true;
                            intcount [j] := intcount[j] + 1;
                         end
                       else j := j - 1;
                    if not already then update;
                  end; {else}
        end; {for}
     writeln ;
     prompt  ('Now writing assonance index file. ');
     assign (textfile, location+'ASSON.NDX');
     rewrite (textfile);
     for i := 1 to count do
          writeln (textfile, asslst[i], intcount[i]:7);
     status (itsa, count);
     close (textfile);
end; {procedure make_assonance_list}

procedure make_rime_list; {generic list routine - called for rimes only
                           because long lists are too slow for suffixes}

var thisone       : rime;            {global to internal routines}

{internal} procedure install (n : wnumber; thislist : list_ptr);
var  endlist : list_ptr;
begin
     {routine to install a list of integers}
     while thislist^.next <> nil do     {step along list to end}
           thislist := thislist^.next;
     new (endlist);                     {new end to the list}
     thislist^.next := endlist;         {set pointer to new end}
     thislist := thislist^.next;        {enter new end entry}
     thislist^.no := n;                 {install the number passed}
end;  {procedure install}

{internal} function member (n : integer; thislist : list_ptr) : wnumber;
var i : integer;
begin
     if n > 1 then
        for i := 2 to n do
          thislist := thislist^.next;
     member := thislist^.no;
end;

begin {procedure make_rime_list proper}
     mark (newheap);
     prompt_string := 'Indexing ' + kindof + ':';
     prompt (prompt_string);
     screenlin := 0;
     count := 0;
     gotoxy (count_x, prompt_y);
     reversv ('  Word #');
     count_x := wherex;
     for i := 0 to number-1 do
       begin
          gotoxy (count_x, prompt_y);
          reversi (i);
          thisone := thisentry[i+1]^.rime;
          if count = 0 then
            update
          else
            begin {outer else}
              j := count;
              already := false;
              while (j >= 1)
                and (not already) do
                 with list_array[j] do
                  if thisone = name then
                     begin
                        already := true;
                        install (i, list);
                        num  := num  + 1;
                     end
          {inner} else
                     j := j - 1;
              if not already then
                 update;
            end; {outer else}
       end; {for}
     assign (textfile, location+filename+'.NDX');
     rewrite (textfile);
     total := 0;
     status (itsa, count);
     prompt ('Now writing first rime index file.');
     for i := 1 to count do
      with list_array[i] do
       begin
          writeln (textfile, name, num:3, total:6);
          total := total + num ;
       end;
     close (textfile);
     prompt ('Now writing second rime index file.');
     assign (ndex, location+filename+'.NDI');
     rewrite (ndex);
     for i := 1 to count do
      with list_array[i] do
        for j := 1 to num do
          begin
            k := member (j, list);
            write (ndex, k);
          end;
     close (ndex);
     release (newheap);
end;  {procedure make_rime_list}

procedure make_suffix_list; {for suffixes only - uses arrays
                             instead of lists for speed}
begin 
     itsa := suf;
     prompt ('Making list of suffixes:');
     screenlin := 0;
     count := 0;
     for i := 0 to number-1 do
       begin
          with thisentry[i+1]^ do
               if count = 0 then
                 update
               else
                 begin {outer else}
                    j := count;
                    already := false;
                    while (j >= 1)
                      and (not already) do
                       if suffix = suflst[j] then
                         begin
                            already := true;
                            intcount[j] := intcount[j] + 1;
                         end
               {inner} else
                         j := j - 1;
                    if not already then
                       update;
                 end; {outer else}
        end; {for}
     prompt ('Now writing first suffix index file.');
     assign (textfile, location+'SUFFIX.NDX');
     rewrite (textfile);
     total := 0;
     for i := 1 to count do
       begin
          writeln (textfile, suflst[i], intcount[i]:6, total:6);
          total := total + intcount[i];
       end; {for}
     close (textfile);
     status (itsa, count);
     prompt ('Indexing: Suffix #');
     k := 1;
     textbackground (white);
     textcolor (black);
     for i := 1 to count do
       begin
          gotoxy (count_x, prompt_y);
          reversi (i);
          for j := 1 to number do
           with thisentry[j]^ do
             if suffix = suflst[i] then
                begin
                   intarray[k] := j - 1;
                   k := k + 1;
                end;
       end; {for}
     prompt ('Now writing second suffix index file.');
     assign (ndex, location+'SUFFIX.NDI');
     rewrite (ndex);
     for i := 1 to number do
          write (ndex, intarray[i]);
     close (ndex);
     gotoxy (count_x, prompt_y);
     reversv ('  Done updating all files.');
     writeln;
end; {procedure make_suffix_list}

{END indexing routines}

begin  {main program}
     trademark;
     load_data;
     make_assonance_list;
     itsa := rim;        { initialize variables for make_rime_list }
     kindof := 'rimes';  { -- necessary because of generic nature  }
     filename := 'RIME'; {    of that routine.                     }
     make_rime_list;
     make_suffix_list; 
     release (heaptop);
end.
